package de.bokeh.icfp2008;

import java.io.*;
import java.util.logging.*;


/**
 * Main class of rover controller.
 */
public class Controller {

    private static final Logger logger = Logger.getLogger("icfp2008");

    public static void main(String[] args) throws IOException {
        if (args.length < 2) {
            System.err.println("usage: Controller host port [options...]");
            return;
        }
        String host = args[0];
        int port = Integer.parseInt(args[1]);

        if (args.length == 2) {
            logger.setLevel(Level.OFF);
        } else {
            logger.setUseParentHandlers(false);
            FileHandler logHandler = new FileHandler("controller.log");
            logHandler.setFormatter(new SimpleFormatter());
            logger.addHandler(logHandler);
            if (args[2].equals("debug"))
                logger.setLevel(Level.ALL);
            else
                logger.setLevel(Level.INFO);
        }

        Controller ctl = new Controller(host, port);
        try {
            ctl.runTrial();
        } catch (Throwable ex) {
            logger.log(Level.SEVERE, "unexpected exception", ex);
        }
    }


    private Protocol protocol;
    private World world;

    Controller(String host, int port) {
        protocol = new Protocol(host, port, logger);
    }

    void runTrial() throws IOException {
        protocol.connect();
        TrialSpecs specs = protocol.receiveTrialSpecs();
        world = new World(specs, logger);
        // Strategy strategy = new RandomStrategy(world);
        Strategy strategy = new ExploreStrategy(world);
        for (int run = 1; ; run++) {
            int score = doRun(strategy);
            if (score == Integer.MIN_VALUE) {
                logger.info("connection closed after run " + (run - 1));
                break;
            }
            logger.info("run " + run + " score: " + score);
        }
    }

    private int doRun(Strategy strategy) throws IOException {
        boolean gotMessage = false;
        AccelerationState[] accel = new AccelerationState[1];
        TurningState[] turn = new TurningState[1];
        
        try {
            for (;;) {
                Message msg = protocol.receiveMessage();
                gotMessage = true;
                switch (msg.kind) {
                case END_OF_RUN:
                    return msg.getScore();
                case CRASH:
                    // immediately followed by telemetry
                    msg = protocol.receiveMessage();
                    break;
                }
                if (msg.kind == Message.Kind.TELEMETRY) {
                    world.setCurrentTelemetry(msg.getTelemetry());
                    strategy.getCommands(accel, turn);
                    protocol.sendCommand(accel[0], turn[0]);
                }
            }
        } catch (EOFException ex) {
            if (!gotMessage)
                return Integer.MIN_VALUE;
            throw ex;
        }
    }

}
