package de.bokeh.icfp2008;

import java.io.*;
import java.net.*;
import java.util.logging.*;


public class Protocol {

    private final Logger logger;
    
    private Socket socket;
    private final String host;
    private final int port;
    private InputStream sin;
    private OutputStream sout;

    public Protocol(String host, int port, Logger logger) {
        this.host = host;
        this.port = port;
        this.logger = logger;
    }


    public void connect() throws IOException {
        socket = new Socket();
        // prefer latency over bandwidth over connect-time.
        socket.setPerformancePreferences(0, 2, 1);
        socket.connect(new InetSocketAddress(host, port));
        socket.setTcpNoDelay(true);
        sin = socket.getInputStream();
        sout = socket.getOutputStream();
        sout.flush();
    }

    public void close() throws IOException {
        socket.close();
        socket = null;
        sin = null;
        sout = null;
    }

    private String receiveNextToken() throws IOException {
        StringBuilder b = new StringBuilder();
        for (;;) {
            int c = sin.read();
            if (c == -1)
                throw new java.io.EOFException();
            if (c == ';') {
                if (b.length() > 0)
                    throw new IOException("semicolon without preceding blank");
                return ";";
            }
            if (c == ' ' || c == '\n') {
                if (b.length() > 0)
                    break;
            } else
                b.append((char) c);
        }
        return b.toString();
    }

    private void expect(String token) throws IOException {
        String t = receiveNextToken();
        if (!t.equals(token))
            throw new IOException("expected '" + token + "' but got '" + t + "'");
    }

    private double receiveFixed() throws IOException {
        String t = receiveNextToken();
        return Double.parseDouble(t);
    }

    private int receiveInt() throws IOException {
        String t = receiveNextToken();
        return Integer.parseInt(t);
    }

    public TrialSpecs receiveTrialSpecs() throws IOException {
        TrialSpecs s = new TrialSpecs();
        expect("I");
        s.dx = receiveFixed();
        s.dy = receiveFixed();
        s.timeLimit = receiveInt();
        s.minSensor = receiveFixed();
        s.maxSensor = receiveFixed();
        s.maxSpeed = receiveFixed();
        s.maxTurn = receiveFixed();
        s.maxHardTurn = receiveFixed();
        expect(";");
        logger.info("received: " + s);
        return s;
    }

    public Message receiveMessage() throws IOException {
        String tok = receiveNextToken();
        int timeStamp = receiveInt();
        Message msg;
        if (tok.equals("T")) {
            String ctl = receiveNextToken();
            if (ctl.length() != 2)
                throw new IOException("vehicle-ctl expected, got '" + ctl + "'");
            AccelerationState accel;
            switch (ctl.charAt(0)) {
            case 'a': accel = AccelerationState.ACCELERATING; break;
            case 'b': accel = AccelerationState.BRAKING; break;
            case '-': accel = AccelerationState.ROLLING; break;
            default: throw new IOException("vehicle-ctl expected, got '" + ctl + "'");
            }
            TurningState turn;
            switch (ctl.charAt(1)) {
            case 'L': turn = TurningState.HARD_LEFT; break;
            case 'l': turn = TurningState.LEFT; break;
            case '-': turn = TurningState.STRAIGHT; break;
            case 'r': turn = TurningState.RIGHT; break;
            case 'R': turn = TurningState.HARD_RIGHT; break;
            default: throw new IOException("vehicle-ctl expected, got '" + ctl + "'");
            }
            Telemetry t = new Telemetry(accel, turn, receiveMoving());
            for (;;) {
                String k = receiveNextToken();
                if (k.equals(";"))
                    break;
                if (k.equals("m")) {
                    t.addMartian(receiveMoving());
                } else {
                    FixedObject.Kind kind;
                    if (k.equals("b"))
                        kind = FixedObject.Kind.BOULDER;
                    else if (k.equals("c"))
                        kind = FixedObject.Kind.CRATER;
                    else if (k.equals("h"))
                        kind = FixedObject.Kind.HOME;
                    else
                        throw new IOException("fixed object expectd, got '" + k + "'");
                    double x = receiveFixed();
                    double y = receiveFixed();
                    double r = receiveFixed();
                    t.addObject(new FixedObject(kind, x, y, r));
                }
            }
            msg = Message.telemetry(timeStamp, t);
        }
        else if (tok.equals("B")) {
            msg = Message.crash(timeStamp);
            expect(";");
        }
        else if (tok.equals("C")) {
            msg = Message.crater(timeStamp);
            expect(";");
        }
        else if (tok.equals("K")) {
            msg = Message.killed(timeStamp);
            expect(";");
        }
        else if (tok.equals("S")) {
            msg = Message.success(timeStamp);
            expect(";");
        }
        else if (tok.equals("E")) {
            int score = receiveInt();
            msg = Message.endOfRun(timeStamp, score);
            expect(";");
        }
        else {
            throw new IOException("Message expected, got '" + tok + "'");
        }
        logger.fine("received: " + msg);
        return msg;
    }

    private MovingObjectState receiveMoving() throws IOException {
        double x = receiveFixed();
        double y = receiveFixed();
        double dir = receiveFixed();
        double speed = receiveFixed();
        return new MovingObjectState(x, y, dir, speed);
    }

    public void sendCommand(AccelerationState a, TurningState t) throws IOException {
        switch (a) {
        case ACCELERATING: sout.write('a'); break;
        case BRAKING: sout.write('b'); break;
        }
        switch (t) {
        case LEFT: sout.write('l'); break;
        case RIGHT: sout.write('r'); break;
        }
        sout.write(';');
        sout.flush();
        logger.fine("sent: " + a + " " + t);
    }
}
