package de.bokeh.icfp2008;

import java.util.*;
import java.util.logging.*;


public class World {

    private final Logger logger;
    
    private final TrialSpecs trialSpecs;
    
    private final Set<FixedObject> craters;
    private final Set<FixedObject> boulders;
    
    private FixedObject home = null;
    
    private Telemetry currentTelemetry = null;
    
    public World(TrialSpecs trialSpecs, Logger logger) {
        this.trialSpecs = trialSpecs;
        this.logger = logger;
        craters = new HashSet<FixedObject>();
        boulders = new HashSet<FixedObject>();
    }

    
    
    public Telemetry getCurrentTelemetry() {
        return currentTelemetry;
    }

    public void setCurrentTelemetry(Telemetry currentTelemetry) {
        this.currentTelemetry = currentTelemetry;
        incorporateTelemetry(currentTelemetry);
    }
    
    private void incorporateTelemetry(Telemetry t) {
        for (FixedObject o : t.getObjects()) {
            switch (o.kind) {
            case BOULDER:
                if (boulders.add(o))
                    logger.info("added boulder to map: " + o);
                break;
            case CRATER:
                if (craters.add(o))
                    logger.info("added crater to map: " + o);
                break;
            case HOME:
                if (!homeIsWhereExpected(o)) {
                    home = o;
                    logger.warning("home not where expected: " + o);
                }
                break;
            default:
                throw new AssertionError(o.kind);
            }
        }
    }

    private static boolean homeIsWhereExpected(FixedObject home) {
        return (home.kind == FixedObject.Kind.HOME
                && home.x == 0.0 && home.y == 0.0
                && home.r == 5.0);
    }

    
    /**
     * @return The current speed as a ratio of the maximum speed; a value between 0 and 1.
     */
    public double getCurrentSpeedRatio() {
        double maxSpeed = trialSpecs.maxSpeed;
        if (maxSpeed == 0.0)
            return 0.0;
        double currentSpeed = currentTelemetry.move.speed;
        return currentSpeed / maxSpeed;
    }

    /**
     * @return <code>true</code> if a crater or martian intersects our movement vector
     * with in 1.5 &times; current speed.
     */
    public boolean isCraterOrMartianAhead() {
        double dangerous = 1.5 * currentTelemetry.move.speed;
        
        for (FixedObject c : craters) {
            double d = Geometry.distance(currentTelemetry.move.x, currentTelemetry.move.y, c.x, c.y);
            d += c.r;
            if (d <= dangerous) {
                // now compute line-circle intersection
                if (Geometry.lineCircleIntersect(currentTelemetry.move.x, currentTelemetry.move.y, currentTelemetry.move.dir,
                        c.x, c.y, c.r))
                    return true;
            }
        }

        for (MovingObjectState c : currentTelemetry.getMartians()) {
            double d = Geometry.distance(currentTelemetry.move.x, currentTelemetry.move.y, c.x, c.y);
            d += 0.8; // give'em some room
            if (d <= dangerous) {
                // now compute line-circle intersection
                if (Geometry.lineCircleIntersect(currentTelemetry.move.x, currentTelemetry.move.y, currentTelemetry.move.dir,
                        c.x, c.y, 0.8))
                    return true;
            }
        }
        
        
        return false;
    }


}
